/**
 * Copyright (C) 2009-2011 enStratusNetworks LLC
 *
 * ====================================================================
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * ====================================================================
 */

package org.dasein.cloud.jclouds.vcloud;

import java.net.URI;
import java.net.URISyntaxException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Locale;
import java.util.Map;

import org.dasein.cloud.CloudException;
import org.dasein.cloud.InternalException;
import org.dasein.cloud.dc.DataCenter;
import org.dasein.cloud.dc.DataCenterServices;
import org.dasein.cloud.dc.Region;
import org.jclouds.rest.RestContext;
import org.jclouds.vcloud.VCloudAsyncClient;
import org.jclouds.vcloud.VCloudClient;
import org.jclouds.vcloud.domain.ReferenceType;
import org.jclouds.vcloud.domain.VDC;

public class VcloudVDC implements DataCenterServices {
    static public final String VDC = "vdc";
    
    private VcloudDirector provider;
    
    VcloudVDC(VcloudDirector cloud) { provider = cloud; }

    @Override
    public DataCenter getDataCenter(String providerDataCenterId) throws InternalException, CloudException {
        RestContext<VCloudClient, VCloudAsyncClient> ctx = provider.getCloudClient();

        try {
            return toDataCenter(ctx, getVDC(providerDataCenterId));
        }
        finally {
            ctx.close();
        }        
    }

    @Override
    public String getProviderTermForDataCenter(Locale locale) {
        return "VDC Unit";
    }

    @Override
    public String getProviderTermForRegion(Locale locale) {
        return "VDC";
    }

    @Override
    public Region getRegion(String providerRegionId) throws InternalException, CloudException {
        for( Region region : listRegions() ) {
            if( region.getProviderRegionId().equals(providerRegionId) ) {
                return region;
            }
        }
        return null;
    }

    public VDC getVDC(String vdcId) {
        RestContext<VCloudClient, VCloudAsyncClient> ctx = provider.getCloudClient();

        try {
            return ctx.getApi().getVDCClient().getVDC(provider.toHref(ctx, vdcId));
        }
        finally {
            ctx.close();
        }
    }
    
    @Override
    public Collection<DataCenter> listDataCenters(String providerRegionId) throws InternalException, CloudException {
        if( providerRegionId != null && !providerRegionId.equals(provider.getContext().getRegionId()) ) {
            return Collections.emptyList();
        }
        RestContext<VCloudClient, VCloudAsyncClient> ctx = provider.getCloudClient();

        try {
            Map<String,ReferenceType> map = provider.getOrg().getVDCs();
            System.out.println("Data centers=" + map);
            ArrayList<DataCenter> dcs = new ArrayList<DataCenter>();
            
            if( map == null ) {
                return Collections.emptyList();
            }
            for( ReferenceType type : map.values() ) {
                VDC vdc = ctx.getApi().getVDCClient().getVDC(type.getHref());
                DataCenter dc = toDataCenter(ctx, vdc);
                
                if( dc != null ) {
                    dcs.add(dc);
                }
            }
            System.out.println("Result=" + dcs);
            return dcs;
        }
        finally {
            ctx.close();
        }
    }

    @Override
    public Collection<Region> listRegions() throws InternalException, CloudException {
        try {
            return Collections.singletonList(toRegion(provider.getContext().getEndpoint()));
        }
        catch( URISyntaxException e ) {
            throw new InternalException("Improper endpoint: " + provider.getContext().getEndpoint());
        }
    }
    
    private Region toRegion(String endpoint) throws URISyntaxException {
        URI uri = new URI(endpoint);
        Region region = new Region();
    
        region.setProviderRegionId(uri.getHost());
        region.setName(region.getProviderRegionId());
        region.setActive(true);
        region.setAvailable(true);
        return region;
    }
    
    private DataCenter toDataCenter(RestContext<VCloudClient, VCloudAsyncClient> ctx, VDC vdc) {
        if( vdc == null ) {
            return null;
        }
        DataCenter dc = new DataCenter();
        
        System.out.println("DC ID=" + vdc.getHref().toASCIIString());
        dc.setProviderDataCenterId(provider.toId(ctx, vdc.getHref()));
        System.out.println("ID=" + dc.getProviderDataCenterId());
        dc.setActive(true);
        dc.setAvailable(true);
        dc.setName(vdc.getName());
        dc.setRegionId(provider.getContext().getRegionId());
        return dc;        
    }

}
